package net.sergeych.lynon

import net.sergeych.bintools.ByteChunk
import kotlin.math.roundToInt

/**
 * LZW compression algorithm: work in progress.
 *
 * Uses Lyng but input/output. Uses automatic code size.
 *
 * TODO: - reset dictionary
 */
class LZW {
    companion object {

        val MAX_CODE_SIZE = 12
        val STOP_CODE = (1 shl MAX_CODE_SIZE) - 1
        val MAX_DICT_SIZE = (STOP_CODE * 0.92).roundToInt()


        /**
         * Compresses the input string using LZW algorithm
         * @param input The string to compress
         * @return List of compressed codes
         */
        fun compress(input: UByteArray,bitOutput: BitOutput) {
            // Initialize dictionary with all possible single characters
            val dictionary = mutableMapOf<ByteChunk, Int>()
            for (i in 0..255) {
                // 23
                dictionary[ByteChunk(ubyteArrayOf(i.toUByte()))] = i
            }

            var nextCode = 256
            var current = ByteChunk(ubyteArrayOf())
//            val result = mutableListOf<Int>()

            for (char in input) {
                val combined = current + char
                if (dictionary.containsKey(combined)) {
                    current = combined
                } else {
                    val size = sizeInBits(dictionary.size)
                    bitOutput.putBits(dictionary[current]!!,size)
                    dictionary[combined] = nextCode++
                    current = ByteChunk(ubyteArrayOf(char))
                }
            }

            if (current.size > 0) {
                val size = sizeInBits(dictionary.size)
                bitOutput.putBits(dictionary[current]!!,size)
            }
        }

        /**
         * Decompresses a list of LZW codes back to the original string
         * @param compressed The list of compressed codes
         * @return The decompressed string
         */
        fun decompress(compressed: BitInput): UByteArray {
            // Initialize dictionary with all possible single characters
            val dictionary = mutableMapOf<Int, UByteArray>()
            for (i in 0..255) {
                dictionary[i] = ubyteArrayOf(i.toUByte())
            }

            var nextCode = 256
            var previous = dictionary[compressed.getBits(9).toInt()]!!
            val result = mutableListOf<UByte>()

            while( !compressed.isEndOfStream ) {
                val codeSize = sizeInBits(nextCode + 1)
                val code = compressed.getBitsOrNull(codeSize)?.toInt() ?: break
                val current = if ( code in dictionary) {
                    dictionary[code]!!
                } else if (code == nextCode) {
                    // Special case for pattern like cScSc
                    previous + previous[0]
                } else {
                    throw IllegalArgumentException("Invalid compressed code: $code")
                }

                result += current
                dictionary[nextCode++] = previous + current[0]
                previous = current
            }

            return result.toTypedArray().toUByteArray()
        }
    }
}

private operator fun ByteChunk.plus(byte: UByte): ByteChunk {
    return ByteChunk(data + byte)
}
